package net.sf.twip;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import net.sf.twip.internal.FailedFrameworkMethod;
import net.sf.twip.internal.FrameworkMethodWithArgs;
import net.sf.twip.internal.TwipConfigurationErrorCantAccessExtension;
import net.sf.twip.internal.TwipConfigurationErrorCantInstantiateExtension;
import net.sf.twip.parameterhandler.AutoTwipParameterHandler;
import net.sf.twip.parameterhandler.BooleanParameterHandler;
import net.sf.twip.parameterhandler.ByteParameterHandler;
import net.sf.twip.parameterhandler.CharacterParameterHandler;
import net.sf.twip.parameterhandler.DoubleParameterHandler;
import net.sf.twip.parameterhandler.EnumParameterHandler;
import net.sf.twip.parameterhandler.FloatParameterHandler;
import net.sf.twip.parameterhandler.IntegerParameterHandler;
import net.sf.twip.parameterhandler.LongParameterHandler;
import net.sf.twip.parameterhandler.ParameterHandler;
import net.sf.twip.parameterhandler.ShortParameterHandler;
import net.sf.twip.parameterhandler.StringParameterHandler;
import net.sf.twip.util.Parameter;
import net.sf.twip.util.ReverseIterable;

import org.junit.Test;
import org.junit.runners.BlockJUnit4ClassRunner;
import org.junit.runners.model.FrameworkMethod;
import org.junit.runners.model.InitializationError;
import org.junit.runners.model.Statement;

/**
 * "Tests with Properties" allows you to simply add parameters to your JUnit
 * tests. TwiP calls your test with all possible combinations of these
 * parameters... or at least some of them in the case of Integers, etc.
 * 
 * @see "http://twip.sf.net"
 */
public class TwiP extends BlockJUnit4ClassRunner {

	@SuppressWarnings("unchecked")
	private static final Class<TwipExtension>[] NO_EXTENSIONS = new Class[0];

	/** Enable the native Java <code>assert</code> keyword for all tests. */
	static {
		ClassLoader.getSystemClassLoader().setDefaultAssertionStatus(true);
	}

	/**
	 * @return the list of {@link FrameworkMethodWithArgs} that need to be
	 *         called for this method.
	 */
	public static List<FrameworkMethodWithArgs> computeParameterizedTestMethods(Method method) {
		List<FrameworkMethodWithArgs> result = new ArrayList<FrameworkMethodWithArgs>();
		result.add(new FrameworkMethodWithArgs(method));
		for (final Parameter parameter : ReverseIterable.create(Parameter.of(method))) {
			final List<FrameworkMethodWithArgs> next = new ArrayList<FrameworkMethodWithArgs>();
			for (final Object value : getParameterHandler(parameter).getParameterValues()) {
				for (final FrameworkMethodWithArgs old : result) {
					next.add(new FrameworkMethodWithArgs(old, value));
				}
			}
			result = next;
		}
		return result;
	}

	private static ParameterHandler getParameterHandler(Parameter parameter) {
		final Class<?> type = parameter.getType();
		if (type.isEnum()) {
			return new EnumParameterHandler(parameter);
		} else if (Boolean.class.isAssignableFrom(type) || Boolean.TYPE.equals(type)) {
			return new BooleanParameterHandler(parameter);
		} else if (Byte.class.isAssignableFrom(type) || Byte.TYPE.equals(type)) {
			return new ByteParameterHandler(parameter);
		} else if (Character.class.isAssignableFrom(type) || Character.TYPE.equals(type)) {
			return new CharacterParameterHandler(parameter);
		} else if (Double.class.isAssignableFrom(type) || Double.TYPE.equals(type)) {
			return new DoubleParameterHandler(parameter);
		} else if (Float.class.isAssignableFrom(type) || Float.TYPE.equals(type)) {
			return new FloatParameterHandler(parameter);
		} else if (Integer.class.isAssignableFrom(type) || Integer.TYPE.equals(type)) {
			return new IntegerParameterHandler(parameter);
		} else if (Long.class.isAssignableFrom(type) || Long.TYPE.equals(type)) {
			return new LongParameterHandler(parameter);
		} else if (Short.class.isAssignableFrom(type) || Short.TYPE.equals(type)) {
			return new ShortParameterHandler(parameter);
		} else if (String.class.isAssignableFrom(type)) {
			return new StringParameterHandler(parameter);
		} else {
			return new AutoTwipParameterHandler(parameter);
		}
	}

	private List<FrameworkMethod> testMethods;

	private final TwipExtension[] extensions;

	public TwiP(Class<?> testClass) throws InitializationError {
		super(testClass);

		final TwipExtensions extensions = testClass.getAnnotation(TwipExtensions.class);
		final Class<? extends TwipExtension>[] classes = (extensions == null) ? NO_EXTENSIONS : extensions
				.value();
		this.extensions = new TwipExtension[classes.length];

		final List<Throwable> errors = new ArrayList<Throwable>();

		for (int i = 0; i < classes.length; i++) {
			try {
				this.extensions[i] = classes[i].newInstance();
			} catch (final InstantiationException e) {
				errors.add(new TwipConfigurationErrorCantInstantiateExtension("can't instantiate "
						+ classes[i], e));
			} catch (final IllegalAccessException e) {
				errors.add(new TwipConfigurationErrorCantAccessExtension("can't access " + classes[i], e));
			}
		}

		if (!errors.isEmpty()) {
			throw new InitializationError(errors);
		}
	}

	@Override
	protected List<FrameworkMethod> computeTestMethods() {
		if (testMethods == null) {
			testMethods = new ArrayList<FrameworkMethod>();
			for (final FrameworkMethod method : getTestClass().getAnnotatedMethods(Test.class)) {
				if (method.getMethod().getParameterTypes().length == 0) {
					testMethods.add(method);
					continue;
				}
				List<? extends FrameworkMethod> parameterizedTestMethods;
				try {
					parameterizedTestMethods = computeParameterizedTestMethods(method.getMethod());
				} catch (final Throwable e) {
					parameterizedTestMethods = Collections.singletonList(new FailedFrameworkMethod(method
							.getMethod(), e));
				}
				testMethods.addAll(parameterizedTestMethods);
			}
		}
		return testMethods;
	}

	@Override
	protected Statement methodInvoker(FrameworkMethod method, final Object testInstance) {
		Statement statement = super.methodInvoker(method, testInstance);
		for (final TwipExtension extension : extensions) {
			statement = extension.wrap(true, method, testInstance, statement);
		}
		return statement;
	}

	@Override
	protected String testName(FrameworkMethod method) {
		return (method instanceof FrameworkMethodWithArgs) //
		? ((FrameworkMethodWithArgs) method).getDescription()
				: super.testName(method);
	}

	@Override
	protected void validatePublicVoidNoArgMethods(Class<? extends Annotation> annotation, boolean isStatic,
			List<Throwable> errors) {
		final List<FrameworkMethod> methods = getTestClass().getAnnotatedMethods(annotation);
		for (final FrameworkMethod eachTestMethod : methods) {
			eachTestMethod.validatePublicVoid(isStatic, errors);
		}
	}

	@Override
	protected Statement withAfters(FrameworkMethod method, final Object testInstance, Statement statement) {
		Statement result = super.withAfters(method, testInstance, statement);
		for (final TwipExtension extension : extensions) {
			result = extension.wrap(false, method, testInstance, result);
		}
		return result;
	}
}
